import os

import numpy as np
import torch
import torch.nn as nn

from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling2
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling6
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling8
from espnet2.asr.frontend.default import DefaultFrontend
from espnet2.layers.global_mvn import GlobalMVN
from espnet2.layers.utterance_mvn import UtteranceMVN

from espnet_onnx.utils.function import make_pad_mask
from ..language_models.lm import Embedding
from ..abs_model import AbsModel


class XformerEncoder(nn.Module, AbsModel):
    def __init__(self, model):
        super().__init__()
        self.embed = Embedding(model.embed)
        self.model = model

    def forward(self, speech, mask):
        if (
            isinstance(self.model.embed, Conv2dSubsampling)
            or isinstance(self.model.embed, Conv2dSubsampling2)
            or isinstance(self.model.embed, Conv2dSubsampling6)
            or isinstance(self.model.embed, Conv2dSubsampling8)
        ):
            xs_pad, mask = self.embed(speech, mask)
            xs_pad = xs_pad[0]
        else:
            xs_pad = self.embed(speech)

        xs_pad, masks = self.model.encoders(xs_pad, mask)
        if isinstance(xs_pad, tuple):
            xs_pad = xs_pad[0]
        if self.model.normalize_before:
            xs_pad = self.model.after_norm(xs_pad)

        olens = masks.squeeze(1).sum(1)
        return xs_pad, olens, None

    def get_output_size(self):
        return self.model.encoders[0].size

    def get_dummy_inputs(self):
        feats = torch.randn(1, 100, 80)
        mask = torch.from_numpy(make_pad_mask(
            np.array([feats.shape[1]]))[:, None, :])
        return (feats, mask)

    def get_input_names(self):
        return ['feats', 'mask']

    def get_output_names(self):
        return ['encoder_out', 'encoder_out_lens']

    def get_dynamic_axes(self):
        return {
            'feats': {
                1: 'feats_length'
            },
            'mask': {
                2: 'mask_length'
            }
        }

    def get_model_config(self, asr_model=None, path=None):
        ret = {}
        ret.update(
            enc_type='XformerEncoder',
            model_path=os.path.join(path, 'encoder.onnx'),
            frontend=self.get_frontend_config(asr_model.frontend),
            do_normalize=asr_model.normalize is not None,
            do_preencoder=asr_model.preencoder is not None,
            do_postencoder=asr_model.postencoder is not None
        )
        if ret['do_normalize']:
            ret.update(normalize=self.get_norm_config(
                asr_model.normalize, path))
        # Currently preencoder, postencoder is not supported.
        # if ret['do_preencoder']:
        #     ret.update(preencoder=get_preenc_config(self.model.preencoder))
        # if ret['do_postencoder']:
        #     ret.update(postencoder=get_postenc_config(self.model.postencoder))
        return ret

    def get_frontend_config(self, frontend):
        # currently only default config is supported.
        assert isinstance(
            frontend, DefaultFrontend), 'Currently only DefaultFrontend is supported.'

        stft_config = dict(
            n_fft=frontend.stft.n_fft,
            win_length=frontend.stft.win_length,
            hop_length=frontend.stft.hop_length,
            window=frontend.stft.window,
            center=frontend.stft.center,
            onesided=frontend.stft.onesided,
            normalized=frontend.stft.normalized,
        )
        logmel_config = frontend.logmel.mel_options
        logmel_config.update(log_base=frontend.logmel.log_base)
        return {
            "stft": stft_config,
            "logmel": logmel_config
        }

    def get_norm_config(self, normalize, path):
        if isinstance(normalize, GlobalMVN):
            return {
                "type": "gmvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
                "stats_file": str(path.parent / 'feats_stats.npz')
            }
        elif isinstance(normalize, UtteranceMVN):
            return {
                "type": "utterance_mvn",
                "norm_means": normalize.norm_means,
                "norm_vars": normalize.norm_vars,
                "eps": normalize.eps,
            }



from typing import Union
from typing import List
from typing import Tuple
from typing import Optional
from pathlib import Path
from typeguard import check_argument_types

import os
import logging
import numpy as np
import librosa
import glob

from espnet_onnx.asr.model.encoder import Encoder
from espnet_onnx.asr.model.decoder import get_decoder
from espnet_onnx.asr.model.lm.seqrnn_lm import SequentialRNNLM
from espnet_onnx.asr.model.lm.transformer_lm import TransformerLM
from espnet_onnx.asr.scorer.ctc_prefix_scorer import CTCPrefixScorer
from espnet_onnx.asr.scorer.length_bonus import LengthBonus
from espnet_onnx.asr.scorer.interface import BatchScorerInterface
from espnet_onnx.asr.beam_search.hyps import Hypothesis
from espnet_onnx.asr.beam_search.beam_search import BeamSearch
from espnet_onnx.asr.beam_search.batch_beam_search import BatchBeamSearch
from espnet_onnx.asr.postprocess.build_tokenizer import build_tokenizer
from espnet_onnx.asr.postprocess.token_id_converter import TokenIDConverter

from espnet_onnx.utils.config import get_config
from espnet_onnx.utils.config import get_tag_config


class Speech2Text:
    """Wrapper class for espnet2.asr.bin.asr_infer.Speech2Text

    """

    def __init__(self,
                 tag_name: str = None,
                 model_dir: Union[Path, str] = None,
                 use_quantized: bool = False,
                 ):
        assert check_argument_types()
        if tag_name is None and model_dir is None:
            raise ValueError('tag_name or model_dir should be defined.')

        if tag_name is not None:
            tag_config = get_tag_config()
            if not tag_name in tag_config.keys():
                raise RuntimeError(f'Model path for tag_name "{tag_name}" is not set on tag_config.yaml.'
                                   + 'You have to export to onnx format with `espnet_onnx.export.asr.export_asr.ModelExport`,'
                                   + 'or have to set exported model path in tag_config.yaml.')
            model_dir = tag_config[tag_name]

        # 1. Build asr model
        config_file = glob.glob(os.path.join(model_dir, 'config.*'))[0]
        config = get_config(config_file)

        if use_quantized and 'quantized_model_path' not in config.encoder.keys():
            # check if quantized model config is defined.
            raise RuntimeError(
                'Configuration for quantized model is not defined.')

        # 2.
        self.encoder = Encoder(config.encoder, use_quantized)
        decoder = get_decoder(config.decoder, config.token,
                              config.transducer, use_quantized)
        ctc = CTCPrefixScorer(config.ctc, config.token.eos, use_quantized)

        scorers = {}
        scorers.update(
            decoder=decoder,
            ctc=ctc,
            length_bonus=LengthBonus(len(config.token.list))
        )

        # 2. Build lm model
        if config.lm.use_lm:
            if config.lm.lm_type == 'SequentialRNNLM':
                scorers.update(
                    lm=SequentialRNNLM(config.lm, use_quantized)
                )
            elif config.lm.lm_type == 'TransformerLM':
                scorers.update(
                    lm=TransformerLM(config.lm, use_quantized)
                )

        # 3. Build ngram model
        # Currently not supported.
        # if config.ngram.use_ngram:
        #     if config.ngram.scorer_type == 'full':
        #         scorers.update(
        #             ngram=NgramFullScorer(
        #                 config.ngram.ngram_file,
        #                 config.token.list
        #             )
        #         )
        #     else:
        #         scorers.update(
        #             ngram=NgramPartialScorer(
        #                 config.ngram.ngram_file,
        #                 config.token.list
        #             )
        #         )

        # 4. Build beam search object
        weights = dict(
            decoder=config.weights.decoder,
            ctc=config.weights.ctc,
            lm=config.weights.lm,
            length_bonus=config.weights.length_bonus,
        )
        if config.transducer.use_transducer_decoder:
            # self.beam_search = BSTransducer(
            #     config.beam_search,
            #     config.transducer,
            #     config.token,
            #     decoder=decoder,
            #     # lm=scorers["lm"] if "lm" in scorers else None,
            #     weights=weights
            # )
            raise ValueError('Transducer is currently not supported.')
        else:
            self.beam_search = BeamSearch(
                config.beam_search,
                config.token,
                scorers=scorers,
                weights=weights,
            )

        non_batch = [
            k
            for k, v in self.beam_search.full_scorers.items()
            if not isinstance(v, BatchScorerInterface)
        ]
        if len(non_batch) == 0:
            self.beam_search.__class__ = BatchBeamSearch
            logging.info("BatchBeamSearch implementation is selected.")
        else:
            logging.warning(
                f"As non-batch scorers {non_batch} are found, "
                f"fall back to non-batch implementation."
            )

        # 5. Build text converter
        if config.tokenizer.token_type is None:
            self.tokenizer = None
        elif config.tokenizer.token_type == 'bpe':
            self.tokenizer = build_tokenizer(
                'bpe', config.tokenizer.bpemodel)
        else:
            self.tokenizer = build_tokenizer(
                token_type=config.tokenizer.token_type)

        self.converter = TokenIDConverter(token_list=config.token.list)
        self.config = config

    def __call__(self, speech: np.ndarray) -> List[
        Tuple[
            Optional[str],
            List[str],
            List[int],
            Union[Hypothesis],
        ]
    ]:
        """Inference
        Args:
            data: Input speech data
        Returns:
            text, token, token_int, hyp
        """
        assert check_argument_types()

        # check dtype
        if speech.dtype != np.float32:
            speech = speech.astype(np.float32)

        # data: (Nsamples,) -> (1, Nsamples)
        speech = speech[np.newaxis, :]
        # lengths: (1,)
        lengths = np.array([speech.shape[1]]).astype(np.int64)

        # b. Forward Encoder
        enc, _ = self.encoder(speech=speech, speech_length=lengths)
        if isinstance(enc, tuple):
            enc = enc[0]
        assert len(enc) == 1, len(enc)

        nbest_hyps = self.beam_search(enc[0])[:1]

        results = []
        for hyp in nbest_hyps:
            # remove sos/eos and get results
            last_pos = -1
            if isinstance(hyp.yseq, list):
                token_int = hyp.yseq[1:last_pos].astype(np.int64)
            else:
                token_int = hyp.yseq[1:last_pos].astype(np.int64).tolist()

            # remove blank symbol id, which is assumed to be 0
            token_int = list(filter(lambda x: x != 0, token_int))

            # Change integer-ids to tokens
            token = self.converter.ids2tokens(token_int)
            # since I add 'blank' before sos for onnx computing
            token = token[1:]

            if self.tokenizer is not None:
                text = self.tokenizer.tokens2text(token)
            else:
                text = None
            results.append((text, token, token_int, hyp))

        return results



from typing import Union
from pathlib import Path
from typeguard import check_argument_types

import os
import glob
from datetime import datetime
import shutil

import numpy as np
import torch
from onnxruntime.quantization import quantize_dynamic

from espnet2.bin.asr_inference import Speech2Text
from espnet2.text.sentencepiece_tokenizer import SentencepiecesTokenizer
from .models import (
    get_encoder,
    get_decoder,
    RNNDecoder,
    PreDecoder,
    CTC,
    LanguageModel
)
from .get_config import (
    get_ngram_config,
    get_beam_config,
    get_token_config,
    get_tokenizer_config
)
from espnet_onnx.utils.config import (
    save_config,
    update_model_path
)


class ModelExport:
    def __init__(self, cache_dir: Union[Path, str] = None):
        assert check_argument_types()
        if cache_dir is None:
            cache_dir = Path.home() / ".cache" / "espnet_onnx"

        self.cache_dir = cache_dir

    def export(self, model: Speech2Text, tag_name: str = None, quantize: bool = False):
        assert check_argument_types()
        if tag_name is None:
            tag_name = datetime.now().strftime("%Y%m%d_%H%M%S")

        base_dir = self.cache_dir / tag_name.replace(' ', '-')
        export_dir = base_dir / 'full'
        export_dir.mkdir(parents=True, exist_ok=True)

        # copy model files
        self._copy_files(model, base_dir)

        model_config = self._create_config(model, export_dir)

        # export encoder
        enc_model = get_encoder(model.asr_model.encoder)
        enc_out_size = enc_model.get_output_size()
        self._export_encoder(enc_model, export_dir)
        model_config.update(encoder=enc_model.get_model_config(
            model.asr_model, export_dir))

        # export decoder
        dec_model = get_decoder(model.asr_model.decoder)
        self._export_decoder(dec_model, enc_out_size, export_dir)
        model_config.update(decoder=dec_model.get_model_config(export_dir))

        # export ctc
        ctc_model = CTC(model.asr_model.ctc.ctc_lo)
        self._export_ctc(ctc_model, enc_out_size, export_dir)
        model_config.update(ctc=ctc_model.get_model_config(export_dir))

        # export lm
        if 'lm' in model.beam_search.full_scorers.keys():
            lm_model = LanguageModel(model.beam_search.full_scorers['lm'])
            self._export_lm(lm_model, export_dir)
            model_config.update(lm=lm_model.get_model_config(export_dir))
        else:
            model_config.update(lm=dict(use_lm=False))

        if quantize:
            quantize_dir = base_dir / 'quantize'
            quantize_dir.mkdir(exist_ok=True)
            qt_config = self._quantize_model(export_dir, quantize_dir)
            for m in qt_config.keys():
                if 'predecoder' in m:
                    model_idx = int(m.split('_')[1])
                    model_config['decoder']['predecoder'][model_idx].update(
                        quantized_model_path=qt_config[m])
                else:
                    model_config[m].update(quantized_model_path=qt_config[m])

        config_name = base_dir / 'config.yaml'
        save_config(model_config, config_name)
        update_model_path(tag_name, base_dir)

    def export_from_pretrained(self, tag_name: str, quantize: bool = False):
        assert check_argument_types()
        model = Speech2Text.from_pretrained(tag_name)
        self.export(model, tag_name, quantize)

    def _create_config(self, model, path):
        ret = {}
        if "ngram" in list(model.beam_search.full_scorers.keys()) \
                + list(model.beam_search.part_scorers.keys()):
            ret.update(ngram=get_ngram_config(model))
        else:
            ret.update(ngram=dict(use_ngram=False))

        # Transducer is currently not supported
        ret.update(transducer=dict(use_transducer_decoder=False))
        ret.update(weights=model.beam_search.weights)
        ret.update(beam_search=get_beam_config(
            model.beam_search, model.minlenratio, model.maxlenratio))
        ret.update(token=get_token_config(model.asr_model))
        ret.update(tokenizer=get_tokenizer_config(model.tokenizer, path))
        return ret

    def _export_encoder(self, model, path):
        file_name = os.path.join(path, 'encoder.onnx')
        torch.onnx.export(
            model,
            model.get_dummy_inputs(),
            file_name,
            verbose=False,
            opset_version=11,
            input_names=model.get_input_names(),
            output_names=model.get_output_names(),
            dynamic_axes=model.get_dynamic_axes()
        )

    def _export_decoder(self, dec_model, enc_size, path):
        file_name = os.path.join(path, 'decoder.onnx')
        torch.onnx.export(
            dec_model,
            dec_model.get_dummy_inputs(enc_size),
            file_name,
            verbose=True,
            opset_version=11,
            input_names=dec_model.get_input_names(),
            output_names=dec_model.get_output_names(),
            dynamic_axes=dec_model.get_dynamic_axes()
        )
        # if decoder is RNNDecoder, then export predecoders
        if isinstance(dec_model, RNNDecoder):
            self._export_predecoder(dec_model, path)

    def _export_predecoder(self, dec_model, path):
        for i, att in enumerate(dec_model.model.att_list):
            att_model = PreDecoder(att)
            file_name = os.path.join(path, 'predecoder_%d.onnx' % i)
            torch.onnx.export(
                att_model,
                att_model.get_dummy_inputs(),
                file_name,
                verbose=True,
                opset_version=11,
                input_names=att_model.get_input_names(),
                output_names=att_model.get_output_names(),
                dynamic_axes=att_model.get_dynamic_axes()
            )

    def _export_ctc(self, ctc_model, enc_size, path):
        file_name = os.path.join(path, 'ctc.onnx')
        torch.onnx.export(
            ctc_model,
            ctc_model.get_dummy_inputs(enc_size),
            file_name,
            verbose=True,
            opset_version=11,
            input_names=ctc_model.get_input_names(),
            output_names=ctc_model.get_output_names(),
            dynamic_axes=ctc_model.get_dynamic_axes()
        )

    def _export_lm(self, lm_model, path):
        file_name = os.path.join(path, 'lm.onnx')
        # export encoder
        torch.onnx.export(
            lm_model,
            lm_model.get_dummy_inputs(),
            file_name,
            verbose=True,
            opset_version=11,
            input_names=lm_model.get_input_names(),
            output_names=lm_model.get_output_names(),
            dynamic_axes=lm_model.get_dynamic_axes()
        )

    def _copy_files(self, model, path):
        # copy stats file
        if model.asr_model.normalize is not None \
                and hasattr(model.asr_model.normalize, 'stats_file'):
            stats_file = model.asr_model.normalize.stats_file
            shutil.copy(stats_file, path)

        # copy bpemodel
        if isinstance(model.tokenizer, SentencepiecesTokenizer):
            bpemodel_file = model.tokenizer.model
            shutil.copy(bpemodel_file, path)

    def _quantize_model(self, model_from, model_to):
        ret = {}
        models = glob.glob(os.path.join(model_from, "*.onnx"))
        for m in models:
            basename = os.path.basename(m).split('.')[0]
            export_file = os.path.join(model_to, basename + '_qt.onnx')
            quantize_dynamic(
                m,
                export_file
            )
            ret[basename] = export_file
            os.remove(os.path.join(model_from, basename + '-opt.onnx'))
        return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2019 Shigeki Karita
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)
# Modified by Masao-Someki
"""Positional Encoding Module."""

import math
import torch


def _pre_hook(
    state_dict,
    prefix,
    local_metadata,
    strict,
    missing_keys,
    unexpected_keys,
    error_msgs,
):
    """Perform pre-hook in load_state_dict for backward compatibility.

    Note:
        We saved self.pe until v.0.5.2 but we have omitted it later.
        Therefore, we remove the item "pe" from `state_dict` for backward compatibility.

    """
    k = prefix + "pe"
    if k in state_dict:
        state_dict.pop(k)


class OnnxPositionalEncoding(torch.nn.Module):
    """Positional encoding.

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_len (int): Maximum input length.
        reverse (bool): Whether to reverse the input position. Only for
        the class LegacyRelPositionalEncoding. We remove it in the current
        class RelPositionalEncoding.
    """

    def __init__(self, model, max_len=512, reverse=False):
        """Construct an PositionalEncoding object."""
        super(OnnxPositionalEncoding, self).__init__()
        self.d_model = model.d_model
        self.reverse = model.reverse
        self.xscale = math.sqrt(self.d_model)
        self.pe = None
        self.extend_pe(torch.tensor(0.0).expand(1, max_len))
        self._register_load_state_dict_pre_hook(_pre_hook)

    def extend_pe(self, x):
        """Reset the positional encodings."""
        if self.pe is not None and self.pe.size(1) >= x.size(1):
            if self.pe.dtype != x.dtype or self.pe.device != x.device:
                self.pe = self.pe.to(dtype=x.dtype, device=x.device)
            return
        pe = torch.zeros(x.size(1), self.d_model)
        if self.reverse:
            position = torch.arange(
                x.size(1) - 1, -1, -1.0, dtype=torch.float32
            ).unsqueeze(1)
        else:
            position = torch.arange(
                0, x.size(1), dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, dtype=torch.float32)
            * -(math.log(10000.0) / self.d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.pe = pe.to(device=x.device, dtype=x.dtype)

    def forward(self, x: torch.Tensor):
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).
        """
        x = x * self.xscale + self.pe[:, : x.size(1)]
        return x


class OnnxScaledPositionalEncoding(OnnxPositionalEncoding):
    """Scaled positional encoding module.

    See Sec. 3.2  https://arxiv.org/abs/1809.08895

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_len (int): Maximum input length.

    """

    def __init__(self, model, max_len=512):
        """Initialize class."""
        super().__init__(model, max_len)
        self.alpha = torch.nn.Parameter(torch.tensor(1.0))

    def reset_parameters(self):
        """Reset parameters."""
        self.alpha.data = torch.tensor(1.0)

    def forward(self, x):
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).

        """
        x = x + self.alpha * self.pe[:, : x.size(1)]
        return x


class OnnxLegacyRelPositionalEncoding(OnnxPositionalEncoding):
    """Relative positional encoding module (old version).

    Details can be found in https://github.com/espnet/espnet/pull/2816.

    See : Appendix B in https://arxiv.org/abs/1901.02860

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_len (int): Maximum input length.

    """

    def __init__(self, model, max_len=512):
        """Initialize class."""
        super().__init__(
            model, max_len,
            reverse=True
        )

    def forward(self, x):
        """Compute positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).
            torch.Tensor: Positional embedding tensor (1, time, `*`).

        """
        x = x * self.xscale
        pos_emb = self.pe[:, : x.size(1)]
        return x, pos_emb


class OnnxRelPositionalEncoding(torch.nn.Module):
    """Relative positional encoding module (new implementation).

    Details can be found in https://github.com/espnet/espnet/pull/2816.

    See : Appendix B in https://arxiv.org/abs/1901.02860

    Args:
        d_model (int): Embedding dimension.
        dropout_rate (float): Dropout rate.
        max_len (int): Maximum input length.

    """

    def __init__(self, model, max_len=512):
        """Construct an PositionalEncoding object."""
        super(OnnxRelPositionalEncoding, self).__init__()
        self.d_model = model.d_model
        self.xscale = math.sqrt(self.d_model)
        self.pe = None
        self.extend_pe(torch.tensor(0.0).expand(1, max_len))

    def extend_pe(self, x):
        """Reset the positional encodings."""
        if self.pe is not None and self.pe.size(1) >= x.size(1) * 2 - 1:
            # self.pe contains both positive and negative parts
            # the length of self.pe is 2 * input_len - 1
            if self.pe.dtype != x.dtype or self.pe.device != x.device:
                self.pe = self.pe.to(dtype=x.dtype, device=x.device)
            return
        # Suppose `i` means to the position of query vecotr and `j` means the
        # position of key vector. We use position relative positions when keys
        # are to the left (i>j) and negative relative positions otherwise (i<j).
        pe_positive = torch.zeros(x.size(1), self.d_model)
        pe_negative = torch.zeros(x.size(1), self.d_model)
        position = torch.arange(0, x.size(1), dtype=torch.float32).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2, dtype=torch.float32)
            * -(math.log(10000.0) / self.d_model)
        )
        pe_positive[:, 0::2] = torch.sin(position * div_term)
        pe_positive[:, 1::2] = torch.cos(position * div_term)
        pe_negative[:, 0::2] = torch.sin(-1 * position * div_term)
        pe_negative[:, 1::2] = torch.cos(-1 * position * div_term)

        # Reserve the order of positive indices and concat both positive and
        # negative indices. This is used to support the shifting trick
        # as in https://arxiv.org/abs/1901.02860
        pe_positive = torch.flip(pe_positive, [0]).unsqueeze(0)
        pe_negative = pe_negative[1:].unsqueeze(0)
        pe = torch.cat([pe_positive, pe_negative], dim=1)
        self.pe = pe.to(device=x.device, dtype=x.dtype)

    def forward(self, x: torch.Tensor):
        """Add positional encoding.

        Args:
            x (torch.Tensor): Input tensor (batch, time, `*`).

        Returns:
            torch.Tensor: Encoded tensor (batch, time, `*`).

        """
        x = x * self.xscale
        pos_emb = self.pe[
            :,
            self.pe.size(1) // 2 - x.size(1) + 1: self.pe.size(1) // 2 + x.size(1),
        ]
        return x, pos_emb


# class StreamPositionalEncoding(torch.nn.Module):
#     """Streaming Positional encoding.

#     Args:
#         d_model (int): Embedding dimension.
#         dropout_rate (float): Dropout rate.
#         max_len (int): Maximum input length.

#     """

#     def __init__(self, d_model, dropout_rate, max_len=5000):
#         """Construct an PositionalEncoding object."""
#         super(StreamPositionalEncoding, self).__init__()
#         self.d_model = d_model
#         self.xscale = math.sqrt(self.d_model)
#         self.dropout = torch.nn.Dropout(p=dropout_rate)
#         self.pe = None
#         self.tmp = torch.tensor(0.0).expand(1, max_len)
#         self.extend_pe(self.tmp.size(1), self.tmp.device, self.tmp.dtype)
#         self._register_load_state_dict_pre_hook(_pre_hook)

#     def extend_pe(self, length, device, dtype):
#         """Reset the positional encodings."""
#         if self.pe is not None:
#             if self.pe.size(1) >= length:
#                 if self.pe.dtype != dtype or self.pe.device != device:
#                     self.pe = self.pe.to(dtype=dtype, device=device)
#                 return
#         pe = torch.zeros(length, self.d_model)
#         position = torch.arange(0, length, dtype=torch.float32).unsqueeze(1)
#         div_term = torch.exp(
#             torch.arange(0, self.d_model, 2, dtype=torch.float32)
#             * -(math.log(10000.0) / self.d_model)
#         )
#         pe[:, 0::2] = torch.sin(position * div_term)
#         pe[:, 1::2] = torch.cos(position * div_term)
#         pe = pe.unsqueeze(0)
#         self.pe = pe.to(device=device, dtype=dtype)

#     def forward(self, x: torch.Tensor, start_idx: int = 0):
#         """Add positional encoding.

#         Args:
#             x (torch.Tensor): Input tensor (batch, time, `*`).

#         Returns:
#             torch.Tensor: Encoded tensor (batch, time, `*`).

#         """
#         self.extend_pe(x.size(1) + start_idx, x.device, x.dtype)
#         x = x * self.xscale + self.pe[:, start_idx : start_idx + x.size(1)]
#         return self.dropout(x)



import torch
import torch.nn as nn

from espnet2.lm.seq_rnn_lm import SequentialRNNLM as espnetSequentialRNNLM
from espnet2.lm.transformer_lm import TransformerLM as espnetTransformerLM

from .language_models.lm import (
    SequentialRNNLM,
    TransformerLM
)
from .abs_model import AbsModel


class LanguageModel(nn.Module, AbsModel):
    def __init__(self, model):
        super().__init__()
        if isinstance(model, espnetSequentialRNNLM):
            self.model = SequentialRNNLM(model)

        elif isinstance(model, espnetTransformerLM):
            self.model = TransformerLM(model)

    def forward(self, *args, **kwargs):
        return self.model.forward(*args, **kwargs)

    def get_dummy_inputs(self, *args, **kwargs):
        return self.model.get_dummy_inputs(*args, **kwargs)

    def get_input_names(self, *args, **kwargs):
        return self.model.get_input_names(*args, **kwargs)

    def get_output_names(self, *args, **kwargs):
        return self.model.get_output_names(*args, **kwargs)

    def get_dynamic_axes(self, *args, **kwargs):
        return self.model.get_dynamix_axes(*args, **kwargs)

    def get_model_config(self, *args, **kwargs):
        return self.model.get_model_config(*args, **kwargs)



import six
import numpy as np


def end_detect(ended_hyps, i, M=3, D_end=np.log(1 * np.exp(-10))):
    """End detection.
    described in Eq. (50) of S. Watanabe et al
    "Hybrid CTC/Attention Architecture for End-to-End Speech Recognition"
    :param ended_hyps:
    :param i:
    :param M:
    :param D_end:
    :return:
    """
    if len(ended_hyps) == 0:
        return False
    count = 0
    best_hyp = sorted(ended_hyps, key=lambda x: x["score"], reverse=True)[0]
    for m in six.moves.range(M):
        # get ended_hyps with their length is i - m
        hyp_length = i - m
        hyps_same_length = [x for x in ended_hyps if len(x["yseq"]) == hyp_length]
        if len(hyps_same_length) > 0:
            best_hyp_same_length = sorted(
                hyps_same_length, key=lambda x: x["score"], reverse=True
            )[0]
            if best_hyp_same_length["score"] - best_hyp["score"] < D_end:
                count += 1

    if count == M:
        return True
    else:
        return False
    


"""Search algorithms for Transducer models."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import onnxruntime

from espnet_onnx.utils.function import is_prefix
from espnet_onnx.utils.function import recombine_hyps
from espnet_onnx.utils.function import select_k_expansions
from espnet_onnx.utils.function import subtract
from espnet_onnx.utils.function import topk
from espnet_onnx.asr.beam_search.hyps import TransducerHypothesis
from espnet_onnx.asr.beam_search.hyps import ExtendedHypothesis


class BeamSearchTransducer:
    """Beam search implementation for Transducer."""

    def __init__(
        self,
        bs_config,
        td_config,
        token_config,
        decoder: AbsDecoder,
        weights: Dict[str, float],
        lm: torch.nn.Module = None,
    ):
        """Initialize Transducer search module.
        Args:
            decoder: Decoder module.
            joint_network: Joint network module.
            beam_size: Beam size.
            lm: LM class.
            lm_weight: LM weight for soft fusion.
            search_type: Search algorithm to use during inference.
            max_sym_exp: Number of maximum symbol expansions at each time step. (TSD)
            u_max: Maximum output sequence length. (ALSD)
            nstep: Number of maximum expansion steps at each time step. (NSC/mAES)
            prefix_alpha: Maximum prefix length in prefix search. (NSC/mAES)
            expansion_beta:
              Number of additional candidates for expanded hypotheses selection. (mAES)
            expansion_gamma: Allowed logp difference for prune-by-value method. (mAES)
            score_norm: Normalize final scores by length. ("default")
            nbest: Number of final hypothesis.
        """
        self.decoder = decoder
        self.joint_network = onnxruntime.InferenceSession(
            td_config.joint_network.model_path
        )
        self.joint_out_names = [
            d.name for d in self.joint_network.get_outputs()]
        self.joint_in_names = [d.name for d in self.joint_network.get_inputs()]

        self.beam_size = bs_config.beam_size
        self.vocab_size = len(token_config.list)

        self.blank_id = token_config.blank

        if self.beam_size <= 1:
            self.search_algorithm = self.greedy_search

        elif td_config.search_type == "default":
            self.search_algorithm = self.default_beam_search

        elif td_config.search_type == "tsd":
            self.max_sym_exp = td_config.max_sym_exp
            self.search_algorithm = self.time_sync_decoding

        elif td_config.search_type == "alsd":
            self.u_max = td_config.u_max
            self.search_algorithm = self.align_length_sync_decoding

        elif td_config.search_type == "nsc":
            self.nstep = td_config.nstep
            self.prefix_alpha = td_config.prefix_alpha
            self.search_algorithm = self.nsc_beam_search

        elif td_config.search_type == "maes":
            self.nstep = td_config.nstep if td_config.nstep > 1 else 2
            self.prefix_alpha = td_config.prefix_alpha
            self.expansion_gamma = td_config.expansion_gamma
            self.expansion_beta = td_config.expansion_beta
            self.search_algorithm = self.modified_adaptive_expansion_search

        else:
            raise NotImplementedError

        self.use_lm = lm is not None
        self.lm = lm
        self.lm_weight = weights['lm']

        self.score_norm = td_config.score_norm
        self.nbest = td_config.nbest

    def __call__(
        self, enc_out: np.ndarray
    ) -> Union[List[TransducerHypothesis], List[ExtendedHypothesis]]:
        """Perform beam search.
        Args:
            enc_out: Encoder output sequence. (T, D_enc)
        Returns:
            nbest_hyps: N-best decoding results
        """
        nbest_hyps = self.search_algorithm(enc_out)
        return nbest_hyps

    def sort_nbest(
        self, hyps: Union[List[TransducerHypothesis], List[ExtendedHypothesis]]
    ) -> Union[List[TransducerHypothesis], List[ExtendedHypothesis]]:
        """Sort hypotheses by score or score given sequence length.
        Args:
            hyps: TransducerHypothesis.
        Return:
            hyps: Sorted hypothesis.
        """
        if self.score_norm:
            hyps.sort(key=lambda x: x.score / len(x.yseq), reverse=True)
        else:
            hyps.sort(key=lambda x: x.score, reverse=True)

        return hyps[: self.nbest]

    def prefix_search(
        self, hyps: List[ExtendedHypothesis], enc_out_t: np.ndarray
    ) -> List[ExtendedHypothesis]:
        """Prefix search for NSC and mAES strategies.
        Based on https://arxiv.org/pdf/1211.3711.pdf
        """
        for j, hyp_j in enumerate(hyps[:-1]):
            for hyp_i in hyps[(j + 1):]:
                curr_id = len(hyp_j.yseq)
                pref_id = len(hyp_i.yseq)

                if (
                    is_prefix(hyp_j.yseq, hyp_i.yseq)
                    and (curr_id - pref_id) <= self.prefix_alpha
                ):
                    logp = self.joint_network.run(
                        self.joint_out_names,
                        {
                            'enc_out': enc_out_t,
                            'dec_out': hyp_i.dec_out[-1]
                        }
                    )
                    curr_score = hyp_i.score + float(logp[hyp_j.yseq[pref_id]])

                    for k in range(pref_id, (curr_id - 1)):
                        logp = self.joint_network.run(
                            self.joint_out_names,
                            {
                                'enc_out': enc_out_t,
                                'dec_out': hyp_i.dec_out[-1]
                            }
                        )

                        curr_score += float(logp[hyp_j.yseq[k + 1]])

                    hyp_j.score = np.logaddexp(hyp_j.score, curr_score)

        return hyps

    def greedy_search(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Greedy search implementation.
        Args:
            enc_out: Encoder output sequence. (T, D_enc)
        Returns:
            hyp: 1-best hypotheses.
        """
        dec_state = self.decoder.init_state(1)

        hyp = TransducerHypothesis(
            score=0.0, yseq=[self.blank_id], dec_state=dec_state)
        cache = {}

        dec_out, state, _ = self.decoder.score(hyp, cache)

        for enc_out_t in enc_out:
            logp = self.joint_network.run(
                self.joint_out_names,
                {
                    'enc_out': enc_out_t,
                    'dec_out': dec_out
                }
            )
            top_logp = np.max(logp, axis=-1)
            pred = np.argmax(logp, axis=-1)

            if pred != self.blank_id:
                hyp.yseq.append(int(pred))
                hyp.score += float(top_logp)

                hyp.dec_state = state

                dec_out, state, _ = self.decoder.score(hyp, cache)

        return [hyp]

    def default_beam_search(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Beam search implementation.
        Modified from https://arxiv.org/pdf/1211.3711.pdf
        Args:
            enc_out: Encoder output sequence. (T, D)
        Returns:
            nbest_hyps: N-best hypothesis.
        """
        beam = min(self.beam_size, self.vocab_size)
        beam_k = min(beam, (self.vocab_size - 1))

        dec_state = self.decoder.init_state(1)

        kept_hyps = [TransducerHypothesis(
            score=0.0, yseq=[self.blank_id], dec_state=dec_state)]
        cache = {}

        for enc_out_t in enc_out:
            hyps = kept_hyps
            kept_hyps = []

            while True:
                max_hyp = max(hyps, key=lambda x: x.score)
                hyps.remove(max_hyp)

                dec_out, state, lm_tokens = self.decoder.score(max_hyp, cache)

                logp = self.joint_network.run(
                    self.joint_out_names,
                    {
                        'enc_out': enc_out_t,
                        'dec_out': dec_out
                    }
                )
                top_k = topk(logp[1:], beam_k)
                kept_hyps.append(
                    TransducerHypothesis(
                        score=(max_hyp.score + float(logp[0:1])),
                        yseq=max_hyp.yseq[:],
                        dec_state=max_hyp.dec_state,
                        lm_state=max_hyp.lm_state,
                    )
                )

                # NOTE: we do not have score function in lm
                if self.use_lm:
                    lm_scores, lm_state = self.lm.score(
                        lm_tokens, max_hyp.lm_state, None
                    )
                else:
                    lm_state = max_hyp.lm_state

                for logp, k in zip(*top_k):
                    score = max_hyp.score + float(logp)

                    if self.use_lm:
                        score += self.lm_weight * lm_scores[k + 1]

                    hyps.append(
                        TransducerHypothesis(
                            score=score,
                            yseq=max_hyp.yseq[:] + [int(k + 1)],
                            dec_state=state,
                            lm_state=lm_state,
                        )
                    )

                hyps_max = float(max(hyps, key=lambda x: x.score).score)
                kept_most_prob = sorted(
                    [hyp for hyp in kept_hyps if hyp.score > hyps_max],
                    key=lambda x: x.score,
                )
                if len(kept_most_prob) >= beam:
                    kept_hyps = kept_most_prob
                    break

        return self.sort_nbest(kept_hyps)

    def time_sync_decoding(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Time synchronous beam search implementation.
        Based on https://ieeexplore.ieee.org/document/9053040
        Args:
            enc_out: Encoder output sequence. (T, D)
        Returns:
            nbest_hyps: N-best hypothesis.
        """
        beam = min(self.beam_size, self.vocab_size)

        beam_state = self.decoder.init_state(beam)

        B = [
            TransducerHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]
        cache = {}

        if self.use_lm:
            B[0].lm_state = self.lm.zero_state()

        for enc_out_t in enc_out:
            A = []
            C = B

            enc_out_t = np.expand_dims(enc_out_t, 0)

            for v in range(self.max_sym_exp):
                D = []

                beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                    C,
                    beam_state,
                    cache,
                    self.use_lm,
                )

                beam_logp = self.joint_network.run(
                    self.joint_out_names,
                    {
                        'enc_out': enc_out_t,
                        'dec_out': beam_dec_out
                    }
                )
                beam_topk = topk(beam_logp[:, 1:], beam)

                seq_A = [h.yseq for h in A]

                for i, hyp in enumerate(C):
                    if hyp.yseq not in seq_A:
                        A.append(
                            TransducerHypothesis(
                                score=(hyp.score + float(beam_logp[i, 0])),
                                yseq=hyp.yseq[:],
                                dec_state=hyp.dec_state,
                                lm_state=hyp.lm_state,
                            )
                        )
                    else:
                        dict_pos = seq_A.index(hyp.yseq)

                        A[dict_pos].score = np.logaddexp(
                            A[dict_pos].score, (hyp.score +
                                                float(beam_logp[i, 0]))
                        )

                if v < (self.max_sym_exp - 1):
                    if self.use_lm:
                        beam_lm_scores, beam_lm_states = self.lm.batch_score(
                            beam_lm_tokens, [c.lm_state for c in C], None
                        )

                    for i, hyp in enumerate(C):
                        for logp, k in zip(beam_topk[0][i], beam_topk[1][i] + 1):
                            new_hyp = TransducerHypothesis(
                                score=(hyp.score + float(logp)),
                                yseq=(hyp.yseq + [int(k)]),
                                dec_state=self.decoder.select_state(
                                    beam_state, i),
                                lm_state=hyp.lm_state,
                            )

                            if self.use_lm:
                                new_hyp.score += self.lm_weight * \
                                    beam_lm_scores[i, k]
                                new_hyp.lm_state = beam_lm_states[i]

                            D.append(new_hyp)

                C = sorted(D, key=lambda x: x.score, reverse=True)[:beam]

            B = sorted(A, key=lambda x: x.score, reverse=True)[:beam]

        return self.sort_nbest(B)

    def align_length_sync_decoding(self, enc_out: np.ndarray) -> List[TransducerHypothesis]:
        """Alignment-length synchronous beam search implementation.
        Based on https://ieeexplore.ieee.org/document/9053040
        Args:
            h: Encoder output sequences. (T, D)
        Returns:
            nbest_hyps: N-best hypothesis.
        """
        beam = min(self.beam_size, self.vocab_size)

        t_max = int(enc_out.shape[0])
        u_max = min(self.u_max, (t_max - 1))

        beam_state = self.decoder.init_state(beam)

        B = [
            TransducerHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]
        final = []
        cache = {}

        if self.use_lm:
            B[0].lm_state = self.lm.zero_state()

        for i in range(t_max + u_max):
            A = []

            B_ = []
            B_enc_out = []
            for hyp in B:
                u = len(hyp.yseq) - 1
                t = i - u

                if t > (t_max - 1):
                    continue

                B_.append(hyp)
                B_enc_out.append((t, enc_out[t]))

            if B_:
                beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                    B_,
                    beam_state,
                    cache,
                    self.use_lm,
                )

                beam_enc_out = np.stack([x[1] for x in B_enc_out])

                beam_logp = self.joint_network.run(
                    self.joint_out_names,
                    {
                        'enc_out': beam_enc_out,
                        'dec_out': beam_dec_out
                    }
                )
                beam_topk = topk(beam_logp[:, 1:], beam)

                if self.use_lm:
                    beam_lm_scores, beam_lm_states = self.lm.batch_score(
                        beam_lm_tokens,
                        [b.lm_state for b in B_],
                        None,
                    )

                for i, hyp in enumerate(B_):
                    new_hyp = TransducerHypothesis(
                        score=(hyp.score + float(beam_logp[i, 0])),
                        yseq=hyp.yseq[:],
                        dec_state=hyp.dec_state,
                        lm_state=hyp.lm_state,
                    )

                    A.append(new_hyp)

                    if B_enc_out[i][0] == (t_max - 1):
                        final.append(new_hyp)

                    for logp, k in zip(beam_topk[0][i], beam_topk[1][i] + 1):
                        new_hyp = TransducerHypothesis(
                            score=(hyp.score + float(logp)),
                            yseq=(hyp.yseq[:] + [int(k)]),
                            dec_state=self.decoder.select_state(beam_state, i),
                            lm_state=hyp.lm_state,
                        )

                        if self.use_lm:
                            new_hyp.score += self.lm_weight * \
                                beam_lm_scores[i, k]
                            new_hyp.lm_state = beam_lm_states[i]

                        A.append(new_hyp)

                B = sorted(A, key=lambda x: x.score, reverse=True)[:beam]
                B = recombine_hyps(B)

        if final:
            return self.sort_nbest(final)
        else:
            return B

    def nsc_beam_search(self, enc_out: np.ndarray) -> List[ExtendedHypothesis]:
        """N-step constrained beam search implementation.
        Based on/Modified from https://arxiv.org/pdf/2002.03577.pdf.
        Please reference ESPnet (b-flo, PR #2444) for any usage outside ESPnet
        until further modifications.
        Args:
            enc_out: Encoder output sequence. (T, D_enc)
        Returns:
            nbest_hyps: N-best hypothesis.
        """
        beam = min(self.beam_size, self.vocab_size)
        beam_k = min(beam, (self.vocab_size - 1))

        beam_state = self.decoder.init_state(beam)

        init_tokens = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]

        cache = {}

        beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
            init_tokens,
            beam_state,
            cache,
            self.use_lm,
        )

        state = self.decoder.select_state(beam_state, 0)

        if self.use_lm:
            beam_lm_scores, beam_lm_states = self.lm.batch_score(
                beam_lm_tokens,
                [i.lm_state for i in init_tokens],
                None,
            )
            lm_state = beam_lm_states[0]
            lm_scores = beam_lm_scores[0]
        else:
            lm_state = None
            lm_scores = None

        kept_hyps = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=state,
                dec_out=[beam_dec_out[0]],
                lm_state=lm_state,
                lm_scores=lm_scores,
            )
        ]

        for enc_out_t in enc_out:
            hyps = self.prefix_search(
                sorted(kept_hyps, key=lambda x: len(x.yseq), reverse=True),
                enc_out_t,
            )
            kept_hyps = []

            beam_enc_out = np.expand_dims(enc_out_t, 0)

            S = []
            V = []
            for n in range(self.nstep):
                beam_dec_out = np.stack([hyp.dec_out[-1] for hyp in hyps])

                beam_logp = self.joint_network.run(
                    self.joint_out_names,
                    {
                        'enc_out': beam_enc_out,
                        'dec_out': beam_dec_out
                    }
                )
                beam_topk = topk(beam_logp[:, 1:], beam_k)

                for i, hyp in enumerate(hyps):
                    S.append(
                        ExtendedHypothesis(
                            yseq=hyp.yseq[:],
                            score=hyp.score + float(beam_logp[i, 0:1]),
                            dec_out=hyp.dec_out[:],
                            dec_state=hyp.dec_state,
                            lm_state=hyp.lm_state,
                            lm_scores=hyp.lm_scores,
                        )
                    )

                    for logp, k in zip(beam_topk[0][i], beam_topk[1][i] + 1):
                        score = hyp.score + float(logp)

                        if self.use_lm:
                            score += self.lm_weight * float(hyp.lm_scores[k])

                        V.append(
                            ExtendedHypothesis(
                                yseq=hyp.yseq[:] + [int(k)],
                                score=score,
                                dec_out=hyp.dec_out[:],
                                dec_state=hyp.dec_state,
                                lm_state=hyp.lm_state,
                                lm_scores=hyp.lm_scores,
                            )
                        )

                V.sort(key=lambda x: x.score, reverse=True)
                V = subtract(V, hyps)[:beam]

                beam_state = self.decoder.create_batch_states(
                    beam_state,
                    [v.dec_state for v in V],
                    [v.yseq for v in V],
                )
                beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                    V,
                    beam_state,
                    cache,
                    self.use_lm,
                )

                if self.use_lm:
                    beam_lm_scores, beam_lm_states = self.lm.batch_score(
                        beam_lm_tokens, [v.lm_state for v in V], None
                    )

                if n < (self.nstep - 1):
                    for i, v in enumerate(V):
                        v.dec_out.append(beam_dec_out[i])

                        v.dec_state = self.decoder.select_state(beam_state, i)

                        if self.use_lm:
                            v.lm_state = beam_lm_states[i]
                            v.lm_scores = beam_lm_scores[i]

                    hyps = V[:]
                else:
                    beam_logp = self.joint_network.run(
                        self.joint_out_names,
                        {
                            'enc_out': beam_enc_out,
                            'dec_out': beam_dec_out
                        }
                    )

                    for i, v in enumerate(V):
                        if self.nstep != 1:
                            v.score += float(beam_logp[i, 0])

                        v.dec_out.append(beam_dec_out[i])

                        v.dec_state = self.decoder.select_state(beam_state, i)

                        if self.use_lm:
                            v.lm_state = beam_lm_states[i]
                            v.lm_scores = beam_lm_scores[i]

            kept_hyps = sorted(
                (S + V), key=lambda x: x.score, reverse=True)[:beam]

        return self.sort_nbest(kept_hyps)

    def modified_adaptive_expansion_search(
        self, enc_out: np.ndarray
    ) -> List[ExtendedHypothesis]:
        """It's the modified Adaptive Expansion Search (mAES) implementation.
        Based on/modified from https://ieeexplore.ieee.org/document/9250505 and NSC.
        Args:
            enc_out: Encoder output sequence. (T, D_enc)
        Returns:
            nbest_hyps: N-best hypothesis.
        """
        beam = min(self.beam_size, self.vocab_size)
        beam_state = self.decoder.init_state(beam)

        init_tokens = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=self.decoder.select_state(beam_state, 0),
            )
        ]

        cache = {}

        beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
            init_tokens,
            beam_state,
            cache,
            self.use_lm,
        )

        state = self.decoder.select_state(beam_state, 0)

        if self.use_lm:
            beam_lm_scores, beam_lm_states = self.lm.batch_score(
                beam_lm_tokens, [i.lm_state for i in init_tokens], None
            )

            lm_state = beam_lm_states[0]
            lm_scores = beam_lm_scores[0]
        else:
            lm_state = None
            lm_scores = None

        kept_hyps = [
            ExtendedHypothesis(
                yseq=[self.blank_id],
                score=0.0,
                dec_state=state,
                dec_out=[beam_dec_out[0]],
                lm_state=lm_state,
                lm_scores=lm_scores,
            )
        ]

        for enc_out_t in enc_out:
            hyps = self.prefix_search(
                sorted(kept_hyps, key=lambda x: len(x.yseq), reverse=True),
                enc_out_t,
            )
            kept_hyps = []

            beam_enc_out = np.expand_dims(enc_out_t, 0)

            list_b = []
            for n in range(self.nstep):
                beam_dec_out = np.stack([h.dec_out[-1] for h in hyps])

                beam_logp = self.joint_network.run(
                    self.joint_out_names,
                    {
                        'enc_out': beam_enc_out,
                        'dec_out': beam_dec_out
                    }
                )
                k_expansions = select_k_expansions(
                    hyps, beam_logp, beam, self.expansion_gamma, self.expansion_beta
                )

                list_exp = []
                for i, hyp in enumerate(hyps):
                    for k, new_score in k_expansions[i]:
                        new_hyp = ExtendedHypothesis(
                            yseq=hyp.yseq[:],
                            score=new_score,
                            dec_out=hyp.dec_out[:],
                            dec_state=hyp.dec_state,
                            lm_state=hyp.lm_state,
                            lm_scores=hyp.lm_scores,
                        )

                        if k == 0:
                            list_b.append(new_hyp)
                        else:
                            new_hyp.yseq.append(int(k))

                            if self.use_lm:
                                new_hyp.score += self.lm_weight * float(
                                    hyp.lm_scores[k]
                                )

                            list_exp.append(new_hyp)

                if not list_exp:
                    kept_hyps = sorted(list_b, key=lambda x: x.score, reverse=True)[
                        :beam
                    ]

                    break
                else:
                    beam_state = self.decoder.create_batch_states(
                        beam_state,
                        [hyp.dec_state for hyp in list_exp],
                        [hyp.yseq for hyp in list_exp],
                    )

                    beam_dec_out, beam_state, beam_lm_tokens = self.decoder.batch_score(
                        list_exp,
                        beam_state,
                        cache,
                        self.use_lm,
                    )

                    if self.use_lm:
                        beam_lm_scores, beam_lm_states = self.lm.batch_score(
                            beam_lm_tokens, [
                                k.lm_state for k in list_exp], None
                        )

                    if n < (self.nstep - 1):
                        for i, hyp in enumerate(list_exp):
                            hyp.dec_out.append(beam_dec_out[i])
                            hyp.dec_state = self.decoder.select_state(
                                beam_state, i)

                            if self.use_lm:
                                hyp.lm_state = beam_lm_states[i]
                                hyp.lm_scores = beam_lm_scores[i]

                        hyps = list_exp[:]
                    else:
                        beam_logp = self.joint_network.run(
                            self.joint_out_names,
                            {
                                'enc_out': beam_enc_out,
                                'dec_out': beam_dec_out
                            }
                        )
                        for i, hyp in enumerate(list_exp):
                            hyp.score += float(beam_logp[i, 0])

                            hyp.dec_out.append(beam_dec_out[i])
                            hyp.dec_state = self.decoder.select_state(
                                beam_state, i)

                            if self.use_lm:
                                hyp.lm_states = beam_lm_states[i]
                                hyp.lm_scores = beam_lm_scores[i]

                        kept_hyps = sorted(
                            list_b + list_exp, key=lambda x: x.score, reverse=True
                        )[:beam]

        return self.sort_nbest(kept_hyps)



import os

import torch
import torch.nn as nn

from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling2
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling6
from espnet.nets.pytorch_backend.transformer.subsampling import Conv2dSubsampling8
from espnet.nets.pytorch_backend.transformer.embedding import (
    PositionalEncoding,  # noqa: H301
    ScaledPositionalEncoding,  # noqa: H301
    RelPositionalEncoding,  # noqa: H301
    LegacyRelPositionalEncoding,  # noqa: H301
)

from espnet_onnx.utils.function import subsequent_mask
from ..abs_model import AbsModel
from .embed import (
    OnnxPositionalEncoding,
    OnnxScaledPositionalEncoding,
    OnnxRelPositionalEncoding,
    OnnxLegacyRelPositionalEncoding
)


def get_pos_emb(pos_emb):
    if isinstance(pos_emb, LegacyRelPositionalEncoding):
        return OnnxLegacyRelPositionalEncoding(pos_emb)
    elif isinstance(pos_emb, ScaledPositionalEncoding):
        return OnnxScaledPositionalEncoding(pos_emb)
    elif isinstance(pos_emb, RelPositionalEncoding):
        return OnnxRelPositionalEncoding(pos_emb)
    elif isinstance(pos_emb, PositionalEncoding):
        return OnnxPositionalEncoding(pos_emb)
    elif isinstance(pos_emb, nn.Sequential) and len(pos_emb) == 0:
        return pos_emb
    else:
        raise ValueError('Embedding model is not supported.')


class Embedding(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
        
        if not isinstance(model, nn.Embedding):
            if (
                isinstance(model, Conv2dSubsampling)
                or isinstance(model, Conv2dSubsampling2)
                or isinstance(model, Conv2dSubsampling6)
                or isinstance(model, Conv2dSubsampling8)
            ):
                self.model.out[-1] = get_pos_emb(model.out[-1])
            else:
                self.model[-1] = get_pos_emb(model[-1])
    
    def forward(self, x, mask=None):
        if mask is None:
            return self.model(x)
        else:
            xs = self.model(x, mask)
            if isinstance(self.model, Conv2dSubsampling):
                return xs, mask[:, :, :-2:2][:, :, :-2:2]
            elif isinstance(self.model, Conv2dSubsampling2):
                return xs, mask[:, :, :-2:2][:, :, :-2:1]
            elif isinstance(self.model, Conv2dSubsampling6):
                return xs, mask[:, :, :-2:2][:, :, :-4:3]
            elif isinstance(self.model, Conv2dSubsampling8):
                return xs, mask[:, :, :-2:2][:, :, :-2:2][:, :, :-2:2]


class SequentialRNNLM(nn.Module, AbsModel):
    def __init__(self, model):
        super().__init__()
        self.encoder = model.encoder
        self.rnn = model.rnn
        self.rnn_type = model.rnn_type
        self.decoder = model.decoder
        self.nlayers = model.nlayers
        self.nhid = model.nhid

    def forward(self, y, hidden1, hidden2=None):
        # batch_score function.
        emb = self.encoder(y)
        if self.rnn_type == 'LSTM':
            output, (hidden1, hidden2) = self.rnn(emb, (hidden1, hidden2))
        else:
            output, hidden1 = self.rnn(emb, hidden1)

        decoded = self.decoder(
            output.contiguous().view(output.size(0) * output.size(1), output.size(2))
        )
        if self.rnn_type == 'LSTM':
            return (
                decoded.view(output.size(0), output.size(1), decoded.size(1)),
                hidden1,
                hidden2
            )
        else:
            return (
                decoded.view(output.size(0), output.size(1), decoded.size(1)),
                hidden1
            )
    
    def get_dummy_inputs(self):
        tgt = torch.LongTensor([0, 1]).unsqueeze(0)
        hidden = torch.randn(self.nlayers, 1, self.nhid)
        if self.rnn_type == 'LSTM':
            return (tgt, hidden, hidden)
        else:
            return (tgt, hidden)
    
    def get_input_names(self):
        if self.rnn_type == 'LSTM':
            return ['x', 'in_hidden1', 'in_hidden2']
        else:
            return ['x', 'in_hidden1']

    def get_output_names(self):
        if self.rnn_type == 'LSTM':
            return ['y', 'out_hidden1', 'out_hidden2']
        else:
            return ['y', 'out_hidden1']

    def get_dynamix_axes(self):
        ret = {
            'x': {
                0: 'x_batch',
                1: 'x_length'
            },
            'y': {
                0: 'y_batch'
            },
            'in_hidden1': {
                1: 'hidden1_batch'
            },
            'out_hidden1': {
                1: 'out_hidden1_batch'
            }
        }
        if self.rnn_type == 'LSTM':
            ret.update({
                'in_hidden2': {
                    1: 'hidden2_batch'
                },
                'out_hidden2': {
                    1: 'out_hidden2_batch'
                }
            })
        return ret
    
    def get_model_config(self, path):
        return {
            "use_lm": True,
            "model_path": os.path.join(path, "lm.onnx"),
            "lm_type": "SequentialRNNLM",
            "rnn_type": self.rnn_type,
            "nhid": self.nhid,
            "nlayers": self.nlayers
        }


class TransformerLM(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.embed = Embedding(model.embed)
        self.encoder = model.encoder
        self.decoder = model.decoder

    def forward(self, y, mask, cache):
        xs = self.embed(y)
        # forward_one_step of Encoder
        if isinstance(self.encoder.embed, Conv2dSubsampling):
            xs, mask = self.encoder.embed(xs, mask)
        else:
            xs = self.encoder.embed(xs)

        new_cache = []
        for c, e in zip(cache, self.encoder.encoders):
            xs, mask = e(xs, mask, cache=c)
            new_cache.append(xs)

        if self.encoder.normalize_before:
            xs = self.encoder.after_norm(xs)

        h = self.decoder(xs[:, -1])
        return h, new_cache

    def get_dummy_inputs(self):
        tgt = torch.LongTensor([0, 1]).unsqueeze(0)
        ys_mask = tgt != 0
        m = torch.from_numpy(subsequent_mask(ys_mask.shape[-1])[None, :])
        mask = ys_mask[None, :] * m
        cache = [
            torch.zeros((1, 1, self.encoder.encoders[0].size))
            for _ in range(len(self.encoder.encoders))
        ]
        return (tgt, mask, cache)
    
    def get_input_names(self):
        return ['tgt', 'tgt_mask'] \
        + ['cache_%d' % i for i in range(len(self.encoder.encoders))]
    
    def get_output_names(self):
        return ['y'] \
        + ['out_cache_%d' % i for i in range(len(self.encoder.encoders))]
        
    def get_dynamix_axes(self):
        ret = {
            'tgt': {
                0: 'tgt_batch',
                1: 'tgt_length'
            },
            'tgt_mask': {
                0: 'tgt_mask_batch',
                1: 'tgt_mask_length',
                2: 'tgt_mask_height'
            }
        }
        ret.update({
            'cache_%d' % d: {
                0: 'cache_%d_batch' % d,
                1: 'cache_%d_length' % d
            }
            for d in range(len(self.encoder.encoders))
        })
        ret.update({
            'out_cache_%d' % d: {
                0: 'out_cache_%d_batch' % d,
                1: 'out_cache_%d_length' % d
            }
            for d in range(len(self.encoder.encoders))
        })
        return ret
    
    def get_model_config(self, path):
        return {
            "use_lm": True,
            "model_path": os.path.join(path, "lm.onnx"),
            "lm_type": "TransformerLM",
            "odim": self.encoder.encoders[0].size,
            "nlayers": len(self.encoder.encoders)
        }
    


from abc import ABC


class AbsModel(ABC):
    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def get_input_names(self):
        raise NotImplementedError

    def get_output_names(self):
        raise NotImplementedError

    def get_dynamix_axes(self):
        return {}

    def get_model_config(self):
        raise NotImplementedError



from espnet_onnx.utils.config import Config
from espnet_onnx.asr.model.decoders.rnn import RNNDecoder
from espnet_onnx.asr.model.decoders.xformer import XformerDecoder


def get_decoder(config, token_config: Config, td_config: Config, use_quantized: bool = False):
    if td_config.use_transducer_decoder:
        raise ValueError('Transducer is currently not supported.')
    else:
        if config.dec_type == 'RNNDecoder':
            return RNNDecoder(config, use_quantized)
        else:
            return XformerDecoder(config, use_quantized)



import os

from espnet2.text.char_tokenizer import CharTokenizer
from espnet2.text.phoneme_tokenizer import PhonemeTokenizer
from espnet2.text.sentencepiece_tokenizer import SentencepiecesTokenizer
from espnet2.text.word_tokenizer import WordTokenizer


def get_ngram_config(model):
    return {
        "use_ngram": True
    }


def get_beam_config(model, minlenratio, maxlenratio):
    return {
        "beam_size": model.beam_size,
        "pre_beam_ratio": model.pre_beam_size / model.beam_size,
        "pre_beam_score_key": model.pre_beam_score_key,
        "maxlenratio": maxlenratio,
        "minlenratio": minlenratio
    }


def get_token_config(model):
    return {
        "sos": model.sos,
        "eos": model.eos,
        "blank": model.blank_id,
        "list": model.token_list
    }


def get_tokenizer_config(model, path):
    if model is None:
        return {}
    elif isinstance(model, SentencepiecesTokenizer):
        model_name = os.path.basename(model.model)
        return {
            "token_type": "bpe",
            "bpemodel": str(path.parent / model_name)
        }
    elif isinstance(model, WordTokenizer):
        return {
            "token_type": "word"
        }
    elif isinstance(model, CharTokenizer):
        return {
            "token_type": "char"
        }
    elif isinstance(model, PhonemeTokenizer):
        return {
            "token_type": "phn"
        }



from .ctc import CTC
from .lm import LanguageModel

# encoder
from espnet2.asr.encoder.rnn_encoder import RNNEncoder as espnetRNNEncoder
from .encoders.rnn import RNNEncoder
from .encoders.xformer import XformerEncoder

# decoder
from espnet2.asr.decoder.rnn_decoder import RNNDecoder as espnetRNNDecoder
from .decoders.rnn import (
    RNNDecoder,
    PreDecoder
)
from .decoders.xformer import XformerDecoder


def get_encoder(model):
    if isinstance(model, espnetRNNEncoder):
        return RNNEncoder(model)
    else:
        return XformerEncoder(model)


def get_decoder(model):
    if isinstance(model, espnetRNNDecoder):
        return RNNDecoder(model)
    else:
        return XformerDecoder(model)
